{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE FunctionalDependencies #-}
module Types.Value
  ( Value(..)
  , Valuable(..)
  , WitnessFor(..)
  ) where

import Prelude hiding (String)

import Types.Values

import Data.Text (Text)

-- Note: We use a wrapper for the String type in order to be able to define the
-- general Valuable [a] instance further down.  Otherwise it would conflict with
-- our Valuable String instance, since the non-wrapped String type is just an
-- alias for [Char].
data Value = String String
           | Text Text
           | Bool Bool
           | Image Image
           | ImageConversionSettings ImageConversionSettings
           | Template Template
           | Empty
           | Tuple (Value, Value)
           | List [Value]
  deriving (Eq, Show)

class Valuable a where
  toValue :: a -> Value
  fromValue :: Value -> a

instance Valuable String where
  toValue = String
  fromValue = \case
    String a -> a
    _ -> error "unexpected"

instance Valuable Text where
  toValue = Text
  fromValue = \case
    Text a -> a
    _ -> error "unexpected"

instance Valuable Bool where
  toValue = Bool
  fromValue = \case
    Bool a -> a
    _ -> error "unexpected"

instance Valuable Image where
  toValue = Image
  fromValue = \case
    Image a -> a
    _ -> error "unexpected"

instance Valuable ImageConversionSettings where
  toValue = ImageConversionSettings
  fromValue = \case
    ImageConversionSettings a -> a
    _ -> error "unexpected"

instance Valuable Template where
  toValue = Template
  fromValue = \case
    Template a -> a
    _ -> error "unexpected"

instance Valuable () where
  toValue () = Empty
  fromValue = \case
    Empty -> ()
    _ -> error "unexpected"

instance (Valuable a, Valuable b) => Valuable (a, b) where
  toValue (a, b) = Tuple (toValue a, toValue b)
  fromValue = \case
    Tuple (va, vb) -> (fromValue va, fromValue vb)
    _ -> error "unexpected"

instance Valuable a => Valuable [a] where
  toValue = List . map toValue
  fromValue = \case
    List a -> map fromValue a
    _ -> error "unexpected"


class Show w => WitnessFor w t | w -> t, t -> w where
  witnessValue :: w

data StringType = StringType deriving Show
instance WitnessFor StringType String where witnessValue = StringType

data TextType = TextType deriving Show
instance WitnessFor TextType Text where witnessValue = TextType

data BoolType = BoolType deriving Show
instance WitnessFor BoolType Bool where witnessValue = BoolType

data ImageType = ImageType deriving Show
instance WitnessFor ImageType Image where witnessValue = ImageType

data ImageConversionSettingsType = ImageConversionSettingsType deriving Show
instance WitnessFor ImageConversionSettingsType ImageConversionSettings where witnessValue = ImageConversionSettingsType

data TemplateType = TemplateType deriving Show
instance WitnessFor TemplateType Template where witnessValue = TemplateType

data EmptyType = EmptyType deriving Show
instance WitnessFor EmptyType () where witnessValue = EmptyType

data TupleType ta tb = TupleType ta tb deriving Show
instance (WitnessFor ta a, WitnessFor tb b) => WitnessFor (TupleType ta tb) (a, b) where witnessValue = TupleType witnessValue witnessValue

data ListType t = ListType t deriving Show
instance WitnessFor t a => WitnessFor (ListType t) [a] where witnessValue = ListType witnessValue
